import os, tqdm
import numpy as np
import matplotlib.pyplot as plt
from astropy.io import fits
from muchbettermoments import quadratic_2d
from astropy.wcs import WCS
from astropy.nddata import Cutout2D
from astropy.utils.data import download_file
import requests
from bs4 import BeautifulSoup
import warnings

from .mast import tic_by_contamination

def use_pointing_model(coords, pointing_model):
    """Applies pointing model to correct the position of star(s) on postcard.

    Parameters
    ----------
    coords : tuple
        (`x`, `y`) position of star(s).

    pointing_model : astropy.table.Table
        pointing_model for ONE cadence.

    Returns
    -------
    coords : tuple
        Corrected position of star(s).
    """
    pointing_model = np.reshape(list(pointing_model), (3,3))
    A = np.column_stack([coords[0], coords[1], np.ones_like(coords[0])])
    fhat = np.dot(A, pointing_model)
    return fhat[:,0:2]


class ffi:
    """This class allows the user to download all full-frame images for a given sector,
         camera, and chip. It also allows the user to create their own pointing model
         based on each cadence for a given combination of sector, camera, and chip.

    No individual user should have to download all of the full-frame images because
         stacked postcards will be available for the user to download from MAST.

    Parameters
    ----------
    sector : int, optional
    camera : int, optional
    chip : int, optional
    """
    def __init__(self, sector=None, camera=None, chip=None):
        self.sector = sector
        self.camera = camera
        self.chip   = chip


    def download_ffis(self, download_dir=None):
        """
        Downloads entire sector of data into FFI download directory.

        Parameters
        ----------
        download_dir : str
            Location where the data files will be stored.
            Defaults to "~/.eleanor/sector_{}/ffis" if `None` is passed.
        """

        def findAllFFIs(ca, ch):
            nonlocal year, days, url
            calFiles, urlPaths = [], []
            for d in days:
                path = '/'.join(str(e) for e in [url, year, d, ca])+'-'+str(ch)+'/'
                for fn in BeautifulSoup(requests.get(path).text, "lxml").find_all('a'):
                    if fn.get('href')[-7::] == 'ic.fits':
                        calFiles.append(fn.get('href'))
                        urlPaths.append(path)
            return calFiles, urlPaths

        if self.sector in np.arange(1,14,1):
            year=2019
        else:
            year=2020
        # Current days available for ETE-6
        days = np.arange(129,158,1)

        # This URL applies to ETE-6 simulated data ONLY
        url = 'https://archive.stsci.edu/missions/tess/ete-6/ffi/'
        files, urlPaths = findAllFFIs(self.camera, self.chip)

        if download_dir is None:
            # Creates hidden .eleanor FFI directory
            ffi_dir = self._fetch_ffi_dir()
        else:
            ffi_dir = download_dir

        files_in_dir = os.listdir(ffi_dir)

        local_paths = []
        for i in range(len(files)):
            if files[i] not in files_in_dir:
                os.system('cd {} && curl -O -L {}'.format(ffi_dir, urlPaths[i]+files[i]))
            local_paths.append(ffi_dir+files[i])
        self.local_paths = np.array(local_paths)
        return


    def _fetch_ffi_dir(self):
        """Returns the default path to the directory where FFIs will be downloaded.

        By default, this method will return "~/.eleanor/sector_{}/ffis" and create
        this directory if it does not exist.  If the directory cannot be
        access or created, then it returns the local directory (".").

        Returns
        -------
        download_dir : str
            Path to location of `ffi_dir` where FFIs will be downloaded
        """
        download_dir = os.path.join(os.path.expanduser('~'), '.eleanor',
                                    'sector_{}'.format(self.sector), 'ffis')
        if os.path.isdir(download_dir):
            return download_dir
        else:
            # if it doesn't exist, make a new cache directory
            try:
                os.mkdir(download_dir)
            # downloads locally if OS error occurs
            except OSError:
                warnings.warn('Warning: unable to create {}. '
                              'Downloading FFIs to the current '
                              'working directory instead.'.format(download_dir))
                download_dir = '.'

        return download_dir


    def sort_by_date(self):
        """Sorts FITS files by start date of observation."""
        dates, time = [], []
        for f in self.local_paths:
            hdu = fits.open(f)
            hdr = hdu[1].header
            dates.append(hdr['DATE-OBS'])
        dates, fns = np.sort(np.array([dates, self.local_paths]))
        self.local_paths = fns
        self.dates = dates
        return


    def build_pointing_model(self, pos_predicted, pos_inferred, outlier_removal=False):
        """Builds an affine transformation to correct the positions of stars
           from a possibly incorrect WCS.

        Parameters
        ----------
        pos_predicted : tuple
            Positions taken straight from the WCS; [[x,y],[x,y],...] format.
        pos_inferred : tuple
            Positions taken using any centroiding method; [[x,y],[x,y],...] format.
        outlier_removal : bool, optional
            Whether to clip 1-sigma outlier frames. Default `False`.

        Returns
        -------
        xhat : np.ndarray
            (3, 3) affine transformation matrix between WCS positions 
            and inferred positions.
        """
        A = np.column_stack([pos_predicted[:,0], pos_predicted[:,1], np.ones_like(pos_predicted[:,0])])
        f = np.column_stack([pos_inferred[:,0], pos_inferred[:,1], np.ones_like(pos_inferred[:,0])])
        if outlier_removal == True:
            dist = np.sqrt(np.sum((A - f)**2, axis=1))
            mean, std = np.nanmean(dist), np.nanstd(dist)
            lim = 1.0
            A = A[dist < mean + lim*std]
            f = f[dist < mean + lim*std]
        ATA = np.dot(A.T, A)
        ATAinv = np.linalg.inv(ATA)
        ATf = np.dot(A.T, f)
        xhat = np.dot(ATAinv, ATf)
        fhat = np.dot(A, xhat)

        return xhat


    def pointing_model_per_cadence(self):
        """Step through build_pointing_model for each cadence."""

        def find_isolated(x, y):
            """Finds the most isolated, least contaminated sources for pointing model."""
            isolated = []
            for i in range(len(x)):
                x_list  = np.delete(x, np.where(x==x[i]))
                y_list  = np.delete(y, np.where(y==y[i]))
                closest = np.sqrt( (x[i]-x_list)**2 + (y[i]-y_list)**2 ).argmin()
                dist    = np.sqrt( (x[i]-x_list[closest])**2+ (y[i]-y_list[closest])**2 )
                if dist > 8.0:
                    isolated.append(i)
            return np.array(isolated)


        def isolated_center(x, y, image):
            """Finds the centroid of each isolated source with quadratic_2d.

            Parameters
            ----------
            x : array-like
                Initial guesses of x positions for sources.
            y : array-like
                Initial guesses of y positions for sources.            
            image : np.ndarray
                FFI flux data.
                
            Returns
            -------
            cenx : array-like
                Controid x coordinates for all good input sources.
            ceny : array-like
                Centroid y coordinates for all good input sources.
            good : list
                Indexes into input arrays corresponding to good sources.
            """
            cenx, ceny, good = [], [], []
            print("Finding isolated centers of sources")
            for i in range(len(x)):
                 if x[i] > 0. and y[i] > 0.:
                     tpf = Cutout2D(image, position=(x[i], y[i]), size=(7,7), mode='partial')
                     origin = tpf.origin_original
                     cen = quadratic_2d(tpf.data)
                     cenx.append(cen[0]+origin[0]); ceny.append(cen[1]+origin[1])
                     good.append(i)
            cenx, ceny = np.array(cenx), np.array(ceny)
            return cenx, ceny, good

        def apply_pointing_model(xy, matrix):
            pointing_model = matrix
            centroid_xs, centroid_ys = [], []
            new_coords = use_pointing_model(xy, pointing_model)
            return np.array(new_coords)


        pm_fn = 'pointingModel_{}_{}-{}.txt'.format(self.sector, self.camera, self.chip)
        with open(pm_fn, 'w') as tf:
            tf.write('0 1 2 3 4 5 6 7 8\n')

        hdu = fits.open(self.local_paths[0])
        hdr = hdu[1].header
        pos = [hdr['CRVAL1'], hdr['CRVAL2']]

        r = 6.0#*np.sqrt(1.2)
        contam = [0.0, 5e-3]
        tmag_lim = 12.5

        t  = tic_by_contamination(pos, r, contam, tmag_lim)

        for fn in self.local_paths:
            print(fn)
            hdu = fits.open(fn)
            hdr = hdu[1].header
            xy = WCS(hdr).all_world2pix(t['ra'], t['dec'], 1)
            # Triple checks the sources are on the FFI
            onFrame = np.where( (xy[0]>10) & (xy[0]<2092-10) & (xy[1]>10) & (xy[1]<2048-10) )[0]
            xy  = np.array([xy[0][onFrame], xy[1][onFrame]])
            iso = find_isolated(xy[0], xy[1])
            xy  = np.array([xy[0][iso], xy[1][iso]])
            cenx, ceny, good = isolated_center(xy[0], xy[1], hdu[1].data)

            # Triple checks there are no nans; Nans make people sad
            no_nans = np.where( (np.isnan(cenx)==False) & (np.isnan(ceny)==False))
            pos_inferred = np.array( [cenx[no_nans], ceny[no_nans]] )
            xy = np.array( [xy[0][no_nans], xy[1][no_nans]] )

            solution = self.build_pointing_model(xy.T, pos_inferred.T)

            xy = apply_pointing_model(xy.T, solution)
            matrix = self.build_pointing_model(xy, pos_inferred.T, outlier_removal=True)

            sol    = np.dot(matrix, solution)
            sol    = sol.flatten()

            with open(pm_fn, 'a') as tf:
                tf.write('{}\n'.format(' '.join(str(e) for e in sol) ) )



import os, sys

from astropy.utils.data import download_file
from astropy.io import fits
import matplotlib.pyplot as plt
from astropy.wcs import WCS
import numpy as np
import warnings
import pandas as pd
import copy
from .mast import crossmatch_by_position

__all__ = ['Postcard']
ELEANORURL = 'http://astro.uchicago.edu/~bmontet/TESS_postcards/'

class Postcard(object):
    """TESS FFI data for one postcard across one sector.
    
    A postcard is an rectangular subsection cut out from the FFIs. 
    It's like a TPF, but bigger. 
    The Postcard object contains a stack of these cutouts from all available 
    FFIs during a given sector of TESS observations.
    
    Parameters
    ----------
    filename : str
        Filename of the downloaded postcard.
    location : str, optional
        Filepath to `filename`.
    
    Attributes
    ----------
    dimensions : tuple
        (`x`, `y`, `time`) dimensions of postcard.
    flux, flux_err : numpy.ndarray
        Arrays of shape `postcard.dimensions` containing flux or error on flux 
        for each pixel.
    time : float
        ?
    header : dict
        Stored header information for postcard file.
    center_radec : tuple
        RA & Dec coordinates of the postcard's central pixel.
    center_xy : tuple
        (`x`, `y`) coordinates corresponding to the location of 
        the postcard's central pixel on the FFI.
    origin_xy : tuple
        (`x`, `y`) coordinates corresponding to the location of 
        the postcard's (0,0) pixel on the FFI.
    """
    def __init__(self, filename, location=None):
        if location is not None:
            self.filename = '{}{}'.format(location, filename)
            self.local_path = copy.copy(self.filename)
            self.hdu = fits.open(self.local_path)
        else:
            if os.path.isdir('./.eleanor/sector_1/postcards')==True:
                self.filename = './.eleanor/sector_1/postcards/{}'.format(filename)
                self.local_path = self.filename
                self.hdu = fits.open(self.local_path)
            else:
                self.filename = '{}{}'.format(ELEANORURL, filename)
                local_path = download_file(self.filename, cache=True)
                self.local_path = local_path
                self.hdu = fits.open(self.local_path)

    def __repr__(self):
        return "eleanor postcard ({})".format(self.filename)

    def plot(self, frame=0, ax=None, scale='linear', **kwargs):
        """Plots a single frame of a postcard.
        
        Parameters
        ----------
        frame : int, optional
            Index of frame. Default 0.
        
        ax : matplotlib.axes.Axes, optional
            Axes on which to plot. Creates a new object by default.
        
        scale : str
            Scaling for colorbar; acceptable inputs are 'linear' or 'log'.
            Default 'linear'.
        
        **kwargs : passed to matplotlib.pyplot.imshow
        
        Returns
        -------
        ax : matplotlib.axes.Axes
        """

        if ax is None:
            _, ax = plt.subplots(figsize=(8, 7))
        if scale is 'log':
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                dat = np.log10(self.flux[:, :, frame])
                dat[~np.isfinite(dat)] = np.nan
        else:
            dat = self.flux[:, :, frame]

        if ('vmin' not in kwargs) & ('vmax' not in kwargs):
            kwargs['vmin'] = np.nanpercentile(dat, 1)
            kwargs['vmax'] = np.nanpercentile(dat, 99)

        im = ax.imshow(dat, **kwargs)
        ax.set_xlabel('Row')
        ax.set_ylabel('Column')
        cbar = plt.colorbar(im, ax=ax)
        if scale == 'log':
            cbar.set_label('log$_{10}$ Flux')
        else:
            cbar.set_label('Flux')

        # Reset the x/y ticks to the position in the ACTUAL FFI.
        xticks = ax.get_xticks() + self.center_xy[0]
        yticks = ax.get_yticks() + self.center_xy[1]
        ax.set_xticklabels(xticks)
        ax.set_yticklabels(yticks)
        return ax

    def find_sources(self):
        """Finds the cataloged sources in the postcard and returns a table.

        Returns
        -------
        result : astropy.table.Table
            All the sources in a postcard with TIC IDs or Gaia IDs.
        """
        result = crossmatch_by_position(self.center_radec, 0.5, 'Mast.Tic.Crossmatch').to_pandas()
        result = result[['MatchID', 'MatchRA', 'MatchDEC', 'pmRA', 'pmDEC', 'Tmag']]
        result.columns = ['TessID', 'RA', 'Dec', 'pmRA', 'pmDEC', 'Tmag']
        return result


    @property
    def header(self):
        return self.hdu[1].header

    @property
    def center_radec(self):
        return(self.header['CEN_RA'], self.header['CEN_DEC'])

    @property
    def center_xy(self):
        return (self.header['CEN_X'],  self.header['CEN_Y'])

    @property
    def origin_xy(self):
        return (self.header['POSTPIX1'], self.header['POSTPIX2'])

    @property
    def flux(self):
        return self.hdu[2].data

    @property
    def dimensions(self):
        return self.flux.shape

    @property
    def flux_err(self):
        return self.hdu[3].data

    @property
    def time(self):
        return (self.hdu[1].data['TSTOP'] + self.hdu[1].data['TSTART'])/2

    @property
    def wcs(self):
        return WCS(self.header)




import numpy as np
import matplotlib.pyplot as plt
from astropy.io import fits
from astropy.wcs import WCS
from .eleanor import data_products
import matplotlib.animation as animation
from astropy.nddata import Cutout2D
from scipy import ndimage

pm_points, centroid = [], []

def animate(i):
    global pm_points, centroid
    for p in pm_points:
        p.remove()
    for c in centroid:
        c.remove()
    pm_points, centroid = [], []
    ax.imshow(tpf[i], origin='lower')#, **kwargs)
    pm_points.append(ax.scatter(pm_x[i], pm_y[i], s=20, c='k'))
    centroid.append(ax.scatter(cen_x[i], cen_y[i], s=20, c='r'))
    time_text.set_text('Frame {}'.format(i))



a = data_products()
fn = 'hlsp_eleanor_tess_ffi_198593129_v1_lc.fits'
hdu = fits.open(fn)
tpf = hdu[0].data
hdr = hdu[0].header
time, lc = hdu[1].data[0], hdu[1].data[1]

pm = a.get_pointing(header=hdr)
center = [hdr['CENTER_X'], hdr['CENTER_Y']]
theta, delx, dely = pm['medT'].data, pm['medX'].data, pm['medY'].data
# Corrects center to be at (4,4) of TPF following the pointing model
corr_x = center[0]*np.cos(theta) - center[1]*np.sin(theta) + delx
corr_y = center[0]*np.sin(theta) + center[1]*np.cos(theta) + dely
pm_x, pm_y = corr_x-np.median(corr_x)+4, corr_y-np.median(corr_y)+4

cen_x, cen_y =[], []
# Tracks centroid
for i in range(len(tpf)):
    tpf_com = Cutout2D(tpf[i], position=(4,4), size=(2,2))
    tpf_com = tpf_com.data
    com = ndimage.measurements.center_of_mass(tpf_com.T - np.median(tpf_com))
    cen_x.append(com[0])
    cen_y.append(com[1])
cen_x, cen_y = cen_x-np.median(cen_x)+4, cen_y-np.median(cen_y)+4

fig, ax = plt.subplots()
# Adds frame count
time_text = ax.text(5.5, -0.25, '', color='white', fontweight='bold')
time_text.set_text('')

Writer = animation.writers['ffmpeg']
writer = Writer(fps=20, metadata=dict(artist='Adina Feinstein'), bitrate=1800)

# Creates animation
ani = animation.FuncAnimation(fig, animate, frames=len(tpf))
# Adds colorbar
plt.colorbar(ax.imshow(tpf[0], origin='lower'), ax=ax)#, **kwargs), ax=ax)
plt.tight_layout()
#plt.show()
ani.save('track_both.mp4'.format(id), writer=writer)



import numpy as np
import matplotlib.pyplot as plt
import matplotlib
import matplotlib.animation as animation
from astropy.io import fits
import os, sys
import imageio
matplotlib.use("Agg")
from scipy import ndimage
from gaiaTIC import ticPositionByID  as ticID
import math
from photutils import CircularAperture, RectangularAperture, aperture_photometry
from lightkurve import KeplerTargetPixelFile as ktpf
import matplotlib.gridspec as gridspec
import matplotlib.patches as patches
from matplotlib.collections import PatchCollection

# --------------------------
# Functions that create apertures
#  of different size & shape
# --------------------------
def square(pos, l, w, theta):
    return RectangularAperture(pos, l, w, theta)

def circle(pos, r):
    return CircularAperture(pos, r)

# --------------------------
# Plots FFI TPF & lightcurve
# --------------------------
def plotAper(ax1, aperture, lc, tpf, color):
    aperture.plot(color=color, alpha=0.5, fill=0.2, ax=ax1)
    return

# --------------------------              
# Extracts lightcurve from custom aperture
# --------------------------              
def customLC(aperture, tpf):
    lc = []
    for f in range(len(tpf.flux)):
        mask = np.zeros_like(tpf.flux[f], dtype=bool)
        for i in range(len(tpf.flux[f])):
            for j in range(len(tpf.flux[f][0])):
                if tpf.flux[f][i][j] <= 0.35 * tpf.flux[f][3][4]:
                    mask[i,j] = True
                else:
                    mask[i,j] = False

        lc.append(aperture_photometry(tpf.flux[f], aperture, mask=mask)['aperture_sum'].data[0])

    return np.array(lc / np.nanmedian(lc))

# ----------------------
#   CREATES ANIMATION   
# ----------------------
def animate(i):
    global scats, text, lines, lc, custLCC, custLCR, ps

    ax1.imshow(tpf.flux[i], origin='lower')#, vmin=40, vmax=150)
    for scat in scats:
        scat.remove()
    for line in lines:
        line.remove()
    for c in ps:
        c.remove()
    scats, lines, ps= [], [], []
    scats.append(ax1.scatter(x[i], y[i], s=16, c='k'))
    time_text.set_text('Frame {}'.format(i))

    lcNorm = lc.flux / np.nanmedian(lc.flux)
#    lines.append(ax.scatter(lc.time[i], custLCC[i], s=20, c='r'))
#    lines.append(ax.scatter(lc.time[i], custLCR[i], s=20, c='k'))

#    circleShape = patches.Circle((x[i],y[i]), 1.5, fill=False, alpha=0.4)
#    rectanShape = patches.Rectangle((x[i]-1.5,y[i]-1.5), 3.0, 3.0, fill=False)
#    p = PatchCollection([rectanShape, circleShape], alpha=0.4)
#    colors = np.linspace(0,1,2)
#    p.set_array(np.array(colors))
#    p.set_edgecolor('face')
#    ps.append(ax1.add_collection(p))


id = str(sys.argv[1])
tpf = ktpf.from_fits('./figures/{}_tpf.fits'.format(id))
lc = tpf.to_lightcurve()

pointing = 'pointingModel_{}-{}.txt'.format(3,3)

theta, delX, delY = np.loadtxt(pointing, usecols=(1,2,3), skiprows=1, unpack=True)

new_id, pos, tmag = ticID(int(id))
x, y, scats, lines = [], [], [], []
ps = []

for i in range(len(tpf.flux)-1):
    if i == 0:
        x.append( 3.0*np.cos(theta[i+1]) - 4.0*np.sin(theta[i+1]) + delX[i+1] )
        y.append( 3.0*np.sin(theta[i+1]) + 4.0*np.cos(theta[i+1]) + delY[i+1] )
    else:
        x.append( x[i-1]*np.cos(theta[i+1]) - y[i-1]*np.sin(theta[i+1]) + delX[i+1] )
        y.append( x[i-1]*np.sin(theta[i+1]) + y[i-1]*np.cos(theta[i+1]) + delY[i+1] )


fig = plt.figure(figsize=(18,5))
spec = gridspec.GridSpec(ncols=3, nrows=1)
ax  = fig.add_subplot(spec[0, 0:2])
ax1 = fig.add_subplot(spec[0, 2])


time_text = ax1.text(6.0, -0.25, '', color='white', fontweight='bold')
time_text.set_text('')

Writer = animation.writers['ffmpeg']
writer = Writer(fps=16, metadata=dict(artist='Adina Feinstein'), bitrate=1800)

ani = animation.FuncAnimation(fig, animate, frames=len(tpf.flux))
plt.title('TIC {}'.format(id), color='black', fontweight='bold', loc='center')

#apertureC = circle([x,y], 1.5)
#apertureR = square([x,y], 3.0, 3.0, 0.0)
#custLCC = customLC(apertureC, tpf)
#custLCR = customLC(apertureR, tpf)
#custLCC = custLCC / np.nanmedian(custLCC)
#custLCR = custLCR / np.nanmedian(custLCR)


# Writes light curve to FITS file
#lcFile  = '{}_lc.fits'.format(id)
#fits.writeto(lcFile, np.array([lc.time, custLCC]))



#ax.plot(lc.time, custLCC, 'r')
#ax.plot(lc.time, custLCR, 'k')
ax.plot(lc.time, lc.flux/np.nanmedian(lc.flux))

x_cen = math.ceil(pos[0])
y_cen = math.ceil(pos[1])

x_ticks = np.arange(x_cen-5, x_cen+4, 1)
y_ticks = np.arange(y_cen-5, y_cen+4, 1)

plt.xticks(np.arange(0,10,1), x_ticks)
plt.yticks(np.arange(0,10,1), y_ticks)
plt.colorbar(plt.imshow(tpf.flux[0]), ax=ax1)#, vmin=40, vmax=150), ax=ax1)
plt.tight_layout()
#plt.show()

ani.save('{}_customAp.mp4'.format(id), writer=writer)




import matplotlib as mpl
import matplotlib.animation as animation
import numpy as np
import bokeh.io
import bokeh.models
from astropy.io import fits

from astropy.wcs import WCS
from .ffi import use_pointing_model


from .mast import *

__all__ = []


class Visualize(object):
    """
    The main class for creating figures, movies, and interactive plots. 
    Allows the user to have a grand ole time playing with their data!

    Parameters
    ----------
    obj : 
        Object must have minimum attributes of 2D array of flux. 
        Will allow for plotting of both postcards & tpfs.
    """

    def __init__(self, object):
        self.obj = object
        self.header = self.obj.header

        if self.obj.tpf is not None:
            self.flux   = self.obj.tpf
            self.center = (np.nanmedian(self.obj.centroid_xs),
                             np.nanmedian(self.obj.centroid_ys))
            self.pointing_model = self.obj.pointing_model
        else:
            self.flux   = self.obj.flux
            self.center = self.obj.center_xy




    def movie(self, output_fn=None, plot_lc=False, pointing_model=False, aperture=False, **kwargs):
        """
        This function allows the user to create a movie of a TPF or postcard
        
        Parameters
        ----------
        output_fn : str, optional 
            Output filename to save the movie to.
        plot_lc : bool, optional 
            If True, plot the light curve and track location along
            light curve with time-dependent movie (Defaults to False).
        pointing_model : bool, optional 
            If True, plot the movement of the pointing_model
            for a given source; only applicable on TPFs (Defaults to False).
        aperture : bool, optional 
            If True, overplot the aperture on the TPF movie.
        **kwargs : 
            Passed to matplotlib.pyplot.imshow.
        
        Returns
        -------
        ani : matplotlib.animation.Animation
            Movie.
        """
        if pointing_model==True:
            if self.obj.centroid_xs is None:
                print("Sorry, you can only track the pointing model on TPFs, not postcards\n",
                      "Please set pointing_model=False or remove when calling .movie()")
            else:
                xs = self.obj.centroid_xs-self.center[0]+4.0
                ys = self.obj.centroid_ys-self.center[1]+4.0


        if 'vmax' not in kwargs:
            kwargs['vmax'] = np.max(self.flux[0])
        if 'vmin' not in kwargs:
            kwargs['vmin'] = np.min(self.flux[0])
        if 'cmap' not in kwargs:
            kwargs['cmap'] = 'viridis' # the superior colormap


        def animate(i):
            nonlocal xs, ys, scats, line
            ax.imshow(self.flux[i], origin='lower', **kwargs)
            if pointing_model==True:
                for c in scats:
                    c.remove()
                scats=[]
                scats.append(ax.scatter(xs[i], ys[i], s=16, c='k', **kwargs))
            if plot_lc==True:
                for l in line:
                    l.remove()
                line=[]
                line.append(ax1.scatter(self.obj.time[i], self.obj.corr_flux[i], c='r', s=16))
            time_text.set_text('Frame {}'.format(i))


        line, scats = [], []

        # Creates 2 subplots if user requests lc to be plotted
        if plot_lc==True:
            fig  = mpl.pyplot.figure(figsize=(18,5))
            spec = mpl.gridspec.GridSpec(ncols=3, nrows=1)
            ax   = fig.add_subplot(spec[0,2])
            ax1  = fig.add_subplot(spec[0,0:2])
            ax1.plot(self.obj.time, self.obj.corr_flux,'k-')
            ax1.set_ylabel('Corrected Flux')
            ax1.set_xlabel('Time')
            ax1.set_xlim([np.min(self.obj.time)-0.05, np.max(self.obj.time)+0.05])
            ax1.set_ylim([np.min(self.obj.corr_flux)-0.05, np.max(self.obj.corr_flux)+0.05])

        # Creates 1 subplot is user just wants a pixel movie
        else:
            fig  = mpl.pyplot.figure()
            spec = mpl.gridspec.GridSpec(ncols=1, nrows=1)
            ax   = fig.add_subplot(spec[0,0])
            ax.set_xlabel('Pixel Column')
            ax.set_ylabel('Pixel Row')

        if aperture==True:
            print("Aperture has not been implemented yet. Sorry friend, we're working on it!")
            return

        # Plots axes in correct (x,y) coordinate space
        if self.obj.centroid_ys is not None:
            mid_x = int(np.round(self.center[0], 0))
            mid_y = int(np.round(self.center[1], 0))
            ax.set_yticklabels([str(e) for e in np.arange(mid_x-4, mid_x+6,1)])
            ax.set_xticklabels([str(e) for e in np.arange(mid_y-4, mid_y+6,1)])

        # Sets text for pixel movie frames
        time_text = ax.text(5.5, -0.25, '', color='white', fontweight='bold')
        time_text.set_text('')

        # Allows movie to be saved as mp4
        Writer = animation.writers['ffmpeg']
        writer = Writer(fps=20, metadata=dict(artist='Adina D. Feinstein'), bitrate=1800)

        ani = animation.FuncAnimation(fig, animate, frames=len(self.flux))

        mpl.pyplot.colorbar(ax.imshow(self.flux[0], **kwargs), ax=ax)

        mpl.pyplot.tight_layout()

        return ani


    def mark_gaia_sources(self):
        """Mark Gaia sources within a given TPF or postcard.
        
        Hover over the points to reveal the source's TIC ID, Gaia ID, Tmag, and Gmag. 
        Also crossmatches with TIC and identifies closest TIC object.
        """

        from bokeh.models import (ColumnDataSource, HoverTool, BasicTicker,
                          Slider, Button, Label, LinearColorMapper, Span,
                          ColorBar)
        from bokeh.plotting import figure, show, output_file
        from bokeh.palettes import Viridis256

        def create_labels(gaia_ra, gaia_dec):
            ticLabel, tmagLabel = np.zeros(len(gaia_id)), np.zeros(len(gaia_id))
            for i in range(len(gaia_ra)):
                tic, tmag, sep = tic_from_coords([gaia_ra[i], gaia_dec[i]])
                print(tic)
                if sep < 1.0:
                    ticLabel[i]  = tic
                    tmagLabel[i] = tmag[0]
            return ticLabel, tmagLabel

        center = [self.header['CEN_RA'], self.header['CEN_DEC']]
        center_xy = WCS(self.header).all_world2pix(center[0], center[1], 1)

        sources    = cone_search(pos=center, r=0.5, service='Mast.Catalogs.GaiaDR2.Cone')
        sources_xy = WCS(self.header).all_world2pix(sources['ra'], sources['dec'], 1)
        new_coords = []
        for i in range(len(sources_xy[0])):
            xy = [sources_xy[0][i], sources_xy[1][i]]
            coords_corr = use_pointing_model(xy, self.pointing_model[0])
            new_coords.append([coords_corr[0][0],coords_corr[0][1]])
        new_coords = np.array(new_coords)

        # Finds sources that lie in the frame
        in_frame = np.where( (new_coords[:,0] >= center_xy[0]-self.obj.dimensions[1]/2.) &
                             (new_coords[:,0] <= center_xy[0]+self.obj.dimensions[1]/2.) &
                             (new_coords[:,1] >= center_xy[1]-self.obj.dimensions[2]/2.) &
                             (new_coords[:,1] <= center_xy[1]+self.obj.dimensions[2]/2.) )
        gaia_pos_x = new_coords[:,1][in_frame] - center_xy[1] + self.obj.dimensions[1]/2.
        gaia_pos_y = new_coords[:,0][in_frame] - center_xy[0] + self.obj.dimensions[2]/2.
        gaia_g_mag = sources['phot_g_mean_mag'][in_frame]
        gaia_id    = sources['source_id'][in_frame]

        ticLabel, tmagLabel = create_labels(sources['ra'][in_frame], sources['dec'][in_frame])

        # Creates hover label
        hover = HoverTool()
        hover.tooltips = [
            ('TIC', ''),
            ('T_mag', ''),
            ('Gaia ID', ''),
            ('G_mag', '')
            ]

        x_range = (-0.5,self.obj.dimensions[1]-0.5)
        y_range = (-0.5,self.obj.dimensions[2]-0.5)
        p = figure(x_range=x_range, y_range=y_range, toolbar_location='above',
                   plot_width=500, plot_height=450)

        color_mapper = LinearColorMapper(palette='Viridis256', low=np.min(self.flux[0]),
                                         high=np.max(self.flux[0]))
        tpf_img = p.image(image=[self.flux[0]], x=-0.5, y=-0.5, dw=self.obj.dimensions[1],
                          dh=self.obj.dimensions[2], color_mapper=color_mapper)

        # Sets the placement of the tick labels
        p.xaxis.ticker = np.arange(0, self.obj.dimensions[1])
        p.yaxis.ticker = np.arange(0, self.obj.dimensions[2])

        # Sets the tick labels
        x_overrides, y_overrides = {}, {}
        x_list = np.arange(int(center_xy[0]-self.obj.dimensions[1]/2),
                           int(center_xy[0]+self.obj.dimensions[1]/2), 1)
        y_list = np.arange(int(center_xy[1]-self.obj.dimensions[2]/2),
                           int(center_xy[1]+self.obj.dimensions[2]/2), 1)

        for i in range(9):
            ind = str(i)
            x_overrides[i] = str(x_list[i])
            y_overrides[i] = str(y_list[i])

        p.xaxis.major_label_overrides = x_overrides
        p.yaxis.major_label_overrides = y_overrides
        p.xaxis.axis_label = 'Pixel Column Number'
        p.yaxis.axis_label = 'Pixel Row Number'

        # Sets the color bar
        color_bar = ColorBar(color_mapper=color_mapper, location=(0,0), border_line_color=None,
                             ticker=BasicTicker(), title='Intensity', title_text_align='left')
        p.add_layout(color_bar, 'right')

        # Adds points onto image
        source = ColumnDataSource(data=dict(x=gaia_pos_x, y=gaia_pos_y,
                                            tic=ticLabel, tmag=tmagLabel,
                                            gaia=gaia_id, gmag=gaia_g_mag))

        s = p.circle('x', 'y', size=8, source=source, line_color=None,
                     selection_color='red', nonselection_fill_alpha=0.0, nonselection_line_alpha=0.0,
                     nonselection_line_color=None, fill_color='black', hover_alpha=0.9, hover_line_color='white')

        # Activates hover feature
        p.add_tools(HoverTool( tooltips=[("TIC ID", "@tic")  , ("TESS Tmag", "@tmag"),
                                         ("Gaia ID", "@gaia"), ("Gaia Gmag", "@gmag")
                                         ], renderers=[s], mode='mouse', point_policy='snap_to_data') )

        show(p)
        return



    def click_aperture(self, path=None):
        """Interactively set aperture."""

        def click_pixels():
            nonlocal tpf
            """Creates a rectangle over a pixel when that pixel is clicked."""
            coords, rectList = [], []

            fig, ax = mpl.pyplot.subplots()
            ax.imshow(tpf[0], origin='lower')

            def onclick(event):
                """Update figure canvas."""
                nonlocal coords, rectList
                x, y = int(np.round(event.xdata,0)), int(np.round(event.ydata,0))

                # Highlights pixel 
                rect = mpl.patches.Rectangle((x-0.5, y-0.5), 1.0, 1.0)
                rect.set_color('white')
                rect.set_alpha(0.4)
                # Adds pixel if not previously clicked
                if [x,y] not in coords:
                    coords.append([x,y])
                    rectList.append(rect)
                    ax.add_patch(rect)
                fig.canvas.draw()
            cid = fig.canvas.mpl_connect('button_press_event', onclick)
            mpl.pyplot.show()
            mpl.pyplot.close()
            return coords, rectList


        def check_pixels():
            nonlocal tpf, coords, rectList
            """Presents a figure for the user to approve of selected pixels."""
            fig, ax = mpl.pyplot.subplots(1)
            ax.imshow(tpf[0], origin='lower')

            # Recreates patches for confirmation
            for i in range(len(coords)):
                x, y = coords[i][0], coords[i][1]
                rect = mpl.patches.Rectangle((x-0.5, y-0.5), 1.0, 1.0)
                rect.set_color('red')
                rect.set_alpha(0.4)
                ax.add_patch(rect)

            # Make Buttons         
            mpl.pyplot.text(-3.5, 5.5, 'Are you happy\nwith this\naperture?', fontsize=8)
            axRadio  = mpl.pyplot.axes([0.05, 0.45, 0.10, 0.15])
            butRadio = mpl.widgets.RadioButtons(axRadio, ('Yes', 'No'), activecolor='red')
            good=True

            # Checks if user is happy with custom aperture
            def get_status(value):
                nonlocal good
                if value == 'Yes':
                    good=True
                else:
                    good=False

            butRadio.on_clicked(get_status)
            mpl.pyplot.show()
            return good

        hdu = self.header
        tpf = self.flux
        coords, rectList = click_pixels()
        check = check_pixels()

        custlc = []
        if check==True:
            if len(coords) == 0:
                print("You have not selected any pixels. No photometry will be completed.")
            else:
                for f in range(len(tpf)):
                    cadence = []
                    for i in range(len(coords)):
                        cadence.append(tpf[f][coords[i][0], coords[i][1]])
                    custlc.append(np.sum(cadence))
                custlc = np.array(custlc) / np.nanmedian(custlc)
                return custlc

        else:
            self.click_aperture()
            
        return



import sys, os, ast
from .source import Source
from .targetdata import TargetData
from .visualize import Visualize

if __name__ == "__main__":
    # Only gets run if it's called from the command line
    # Takes arguments, downloads TPF to machine
    if len(sys.argv) > 1:
        args = sys.argv

        for a in args[1::]:
            if a[0:2]=='fn':
                star = Source(fn=a[3::])
            elif a[0:6]=='coords':
                star = Source(coords=ast.literal_eval(a[7::]))
            elif a[0:3]=='gaia':
                star = Source(gaia=a[4::])
            elif a[0:3]=='tic':
                star = Source(tic=a[4::])
        print(star.tess_mag)
        data = TargetData(star)
        data.save()
    else:
        raise ValueError("To run this script, please add one of the following "
                         "arguments to your command: tic={int}, gaia={int}, "
                         "coords={list}, fn={str}.")



import numpy as np
import matplotlib.pyplot as plt
from astropy.nddata import Cutout2D
from photutils import CircularAperture, RectangularAperture, aperture_photometry
from lightkurve import SFFCorrector
from scipy.optimize import minimize
from astropy.table import Table, Column
from astropy.wcs import WCS
from time import strftime
from astropy.io import fits
from muchbettermoments import quadratic_2d
import urllib
import os

from .ffi import use_pointing_model
from .postcard import Postcard

__all__  = ['TargetData']

class TargetData(object):
    """
    Object containing the light curve, target pixel file, and related information
    for any given source.

    Parameters
    ----------
    source : ellie.Source
        The source object to use.
    height : int, optional
        Height in pixels of TPF to retrieve. Default 9.
    width : int, optional
        Width in pixels of TPF to retrieve. Default 9.


    Attributes
    ----------
    header : dict
        FITS header for saving/loading data.
    source_info : eleanor.Source
        Pointer to input source.
    aperture : 
        Aperture to use if overriding default. To use default, set to `None`.
    tpf : np.ndarray
        Target pixel file of fluxes; array with shape `dimensions`.
    time : np.ndarray
        Time series.
    post_obj : eleanor.Postcard
        Pointer to Postcard objects containing this TPF.
    pointing_model : astropy.table.Table
        Table of matrices describing the transformation matrix from FFI default
        WCS and eleanor's corrected pointing.
    tpf_err : np.ndarray
        Errors on fluxes in `tpf`.
    centroid_xs : np.ndarray
        Position of the source in `x` inferred from pointing model; has same length as `time`.
        Position is relative to the pixel coordinate system of the postcard.
    centroid_ys : np.ndarray
        Position of the source in `y` inferred from pointing model; has same length as `time`.
        Position is relative to the pixel coordinate system of the postcard.
    cen_x : int
        Median `x` position of the source.
        Position is relative to the pixel coordinate system of the postcard.
    cen_y : int
        Median `y` position of the source.
        Position is relative to the pixel coordinate system of the postcard.
    dimensions : tuple
        Shape of `tpf`. Should be (`time`, `height`, `width`).
    all_apertures : list
        List of aperture objects.
    aperture : array-like
        Chosen aperture for producing `raw_flux` lightcurve. Format is array
        with shape (`height`, `width`). All entries are floats in range [0,1].
    all_lc_err : np.ndarray
        Estimated uncertainties on `all_raw_lc`.
    all_raw_lc : np.ndarray
        All lightcurves extracted using `all_apertures`.
        Has shape (N_apertures, N_time).
    all_corr_lc : np.ndarray
        All systematics-corrected lightcurves. See `all_raw_lc`.
    best_ind : int
        Index into `all_apertures` producing the best (least noisy) lightcurve.
    corr_flux : np.ndarray
        Systematics-corrected version of `raw_flux`.
    flux_err : np.ndarray
        Estimated uncertainty on `raw_flux`.
    raw_flux : np.ndarray
        Un-systematics-corrected lightcurve derived using `aperture` and `tpf`.
    x_com : np.ndarray
        Position of the source in `x` inferred from TPF; has same length as `time`.
        Position is relative to the pixel coordinate system of the TPF.
    y_com : np.ndarray
        Position of the source in `y` inferred from TPF; has same length as `time`.
        Position is relative to the pixel coordinate system of the TPF.
    quality : int
        Quality flag.

    Notes
    -----
    `save()` and `load()` methods write/read these data to a FITS file with format:

    Extension[0] = header
    
    Extension[1] = (N_time, height, width) TPF, where n is the number of cadences in an observing run
    
    Extension[2] = (3, N_time) time, raw flux, systematics corrected flux
    """

    def __init__(self, source, height=9, width=9, save_postcard=True):
        self.source_info = source

        if source.premade:
            self.load()

        else:
            self.aperture = None
            self.post_obj = Postcard(source.postcard)
            self.time  = self.post_obj.time
            self.load_pointing_model(source.sector, source.camera, source.chip)
            self.get_tpf_from_postcard(source.coords, source.postcard, height, width, save_postcard)
            self.create_apertures(height, width)
            self.get_lightcurve()
            self.center_of_mass()
            self.set_quality()
            self.set_header()


    def load_pointing_model(self, sector, camera, chip):

        pointing_link = urllib.request.urlopen('http://jet.uchicago.edu/tess_postcards/pointingModel_{}_{}-{}.txt'.format(sector,
                                                                                                                          camera,
                                                                                                                          chip))
        pointing = pointing_link.read().decode('utf-8')
        pointing = Table.read(pointing, format='ascii.basic') # guide to postcard locations
        self.pointing_model = pointing
        return


    def get_tpf_from_postcard(self, pos, postcard, height, width, save_postcard):
        """Gets TPF from postcard."""

        self.tpf = None
        self.centroid_xs = None
        self.centroid_ys = None

        xy = WCS(self.post_obj.header).all_world2pix(pos[0], pos[1], 1)
        
        # Apply the pointing model to each cadence to find the centroids
        centroid_xs, centroid_ys = [], []
        for i in range(len(self.pointing_model)):
            new_coords = use_pointing_model(xy, self.pointing_model[i])
            centroid_xs.append(new_coords[0][0])
            centroid_ys.append(new_coords[0][1])
        self.centroid_xs = np.array(centroid_xs)
        self.centroid_ys = np.array(centroid_ys)

        # Define tpf as region of postcard around target
        med_x, med_y = np.nanmedian(self.centroid_xs), np.nanmedian(self.centroid_ys)

        med_x, med_y = int(np.round(med_x,0)), int(np.round(med_y,0))

        post_flux = np.transpose(self.post_obj.flux, (2,0,1))
        post_err  = np.transpose(self.post_obj.flux_err, (2,0,1))

        self.cen_x, self.cen_y = med_x, med_y

        y_length, x_length = int(np.floor(height/2.)), int(np.floor(width/2.))

        y_low_lim = med_y-y_length
        y_upp_lim = med_y+y_length+1
        x_low_lim = med_x-x_length
        x_upp_lim = med_x+x_length+1

        post_y_upp, post_x_upp = self.post_obj.dimensions[0], self.post_obj.dimensions[1]

        # Fixes the postage stamp if the user requests a size that is too big for the postcard
        if y_low_lim <= 0:
            y_low_lim = 0
        if x_low_lim <= 0:
            x_low_lim = 0
        if y_upp_lim  > post_y_upp:
            y_upp_lim = post_y_upp+1
        if x_upp_lim >  post_x_upp:
            x_upp_lim = post_x_upp+1

        if (x_low_lim==0) or (y_low_lim==0) or (x_upp_lim==post_x_upp) or (y_upp_lim==post_y_upp):
            print("The size postage stamp you are requesting falls off the edge of the postcard.")
            print("WARNING: Your postage stamp may not be centered.")

        self.tpf     = post_flux[:, y_low_lim:y_upp_lim, x_low_lim:x_upp_lim]
        self.tpf_err = post_err[: , y_low_lim:y_upp_lim, x_low_lim:x_upp_lim]
        self.dimensions = np.shape(self.tpf)

        if save_postcard == False:
            try:
                os.remove(source.postcard.filename)
            except OSError:
                pass
        return


    def create_apertures(self, height, width):
        """Creates a range of sizes and shapes of apertures to test."""
        import eleanor
        self.all_apertures = None

        # Saves some time by pre-loading apertures for 9x9 TPFs
        if (height,width) == (9,9):
            ap_path = eleanor.__path__[0]+'/default_apertures.npy'
            self.all_apertures = np.load(ap_path)

        # Creates aperture based on the requested size of TPF
        else:
            # Creates a circular and rectangular aperture
            def circle(pos,r):
                return CircularAperture(pos,r)
            def rectangle(pos, l, w, t):
                return RectangularAperture(pos, l, w, t)

            # Completes either binary or weighted aperture photometry
            def binary(data, apertures, err):
                return aperture_photometry(data, apertures, error=err, method='center')
            def weighted(data, apertures, err):
                return aperture_photometry(data, apertures, error=err, method='exact')

            r_list = np.arange(1.5,4,0.5)

            # Center gives binary mask; exact gives weighted mask
            circles, rectangles, self.all_apertures = [], [], []

            center = (width/2, height/2)

            for r in r_list:
                ap_circ = circle( center, r )
                ap_rect = rectangle( center, r, r, 0.0)
                circles.append(ap_circ); rectangles.append(ap_rect)
                for method in ['center', 'exact']:
                    circ_mask = ap_circ.to_mask(method=method)[0].to_image(shape=((
                                np.shape( self.tpf[0]))))
                    rect_mask = ap_rect.to_mask(method=method)[0].to_image(shape=((
                                np.shape( self.tpf[0]))))
                    self.all_apertures.append(circ_mask)
                    self.all_apertures.append(rect_mask)
            self.all_apertures = np.array(self.all_apertures)


    def get_lightcurve(self, aperture=False):
        """Extracts a light curve using the given aperture and TPF.

        Allows the user to pass in a mask to use, otherwise sets best lightcurve and aperture (min std).
        Mask is a 2D array of the same shape as TPF (9x9).

        Parameters
        ----------
        aperture : numpy.ndarray
            (`height`, `width`) array of floats in the range [0,1] with desired weights for each pixel to
            create a light curve. If not set, ideal aperture is inferred automatically. If set, uses this
            aperture at the expense of all other set apertures.
        """
        def apply_mask(mask):
            lc     = np.zeros(len(self.tpf))
            lc_err = np.zeros(len(self.tpf))
            for cad in range(len(self.tpf)):
                lc[cad]     = np.sum( self.tpf[cad] * mask)
                lc_err[cad] = np.sqrt( np.sum( self.tpf_err[cad]**2 * mask))
            self.raw_flux   = np.array(lc)
            self.corr_flux  = self.jitter_corr(flux=lc)
            self.flux_err   = np.array(lc_err)
            return


        self.flux_err   = None

        if (self.aperture is None):

            self.all_lc_err  = None

            all_raw_lc     = np.zeros((len(self.all_apertures), len(self.tpf)))
            all_lc_err = np.zeros((len(self.all_apertures), len(self.tpf)))
            all_corr_lc = np.copy(all_raw_lc)

            stds = []
            for a in range(len(self.all_apertures)):
                for cad in range(len(self.tpf)):
                    all_lc_err[a][cad] = np.sqrt( np.sum( self.tpf_err[cad]**2 * self.all_apertures[a] ))
                    all_raw_lc[a][cad] = np.sum( self.tpf[cad] * self.all_apertures[a] )

                all_corr_lc[a] = self.jitter_corr(flux=all_raw_lc[a]/np.nanmedian(all_raw_lc[a]))*np.nanmedian(all_raw_lc[a])
                stds.append( np.std(all_corr_lc[a]) )

            self.all_raw_lc  = np.array(all_raw_lc)
            self.all_lc_err  = np.array(all_lc_err)
            self.all_corr_lc = np.array(all_corr_lc)

            best_ind = np.where(stds == np.min(stds))[0][0]
            self.best_ind = best_ind
            self.corr_flux= self.all_corr_lc[best_ind]
            self.raw_flux = self.all_raw_lc[best_ind]
            self.aperture = self.all_apertures[best_ind]
            self.flux_err = self.all_lc_err[best_ind]

        else:
            if np.shape(aperture) == np.shape(self.tpf[0]):
                self.aperture = aperture
                apply_mask(self.aperture)
            elif self.aperture is not None:
                apply_mask(self.aperture)
            else:
                print("We could not find a custom aperture. Please either create a 2D array that is the same shape as the TPF.")
                print("Or, create a custom aperture using the function TargetData.custom_aperture(). See documentation for inputs.")

        return


    def center_of_mass(self):
        """
        Calculates the position of the source across all cadences using `muchbettermoments` and `self.best_aperture`.

        Finds the brightest pixel in a (`height`, `width`) region summed up over all cadence.
        Searches a smaller (3x3) region around this pixel at each cadence and uses `muchbettermoments` to find the maximum.
        """

        self.x_com = []
        self.y_com = []

        summed_pixels = np.sum(self.aperture * self.tpf, axis=0)
        brightest = np.where(summed_pixels == np.max(summed_pixels))
        cen = (brightest[0][0], brightest[1][0])
        for a in range(len(self.tpf)):
            data = self.tpf[a, cen[0]-3:cen[0]+2, cen[1]-3:cen[1]+2]
            c_0  = quadratic_2d(data)
            c_frame = [cen[0]+c_0[0], cen[1]+c_0[1]]
            self.x_com.append(c_frame[0])
            self.y_com.append(c_frame[1])
        return



    def set_quality(self):
        """Currently (10/13/2018), this function sets a flag for when the centroid is
        3 sigma away from the mean either in the x or y direction.
        Hopefully in the future, MAST will put in some quality flags for us.
        Our flags and their flags will be combnied, if they create flags.
        """
        bad = np.where( (self.centroid_xs > np.mean(self.centroid_xs)+3*np.std(self.centroid_xs)) | (self.centroid_ys > np.mean(self.centroid_ys)+3*np.std(self.centroid_ys)))

        quality = np.zeros(np.shape(self.time))
        quality[bad] = 1
        self.quality = quality



    def psf_lightcurve(self, nstars=1, model='gaussian', xc=[4.5], yc=[4.5]):
        """
        Performs PSF photometry for a selection of stars on a TPF. 
        
        Parameters
        ----------
        nstars: int, optional
            Number of stars to be modeled on the TPF.
        model: string, optional 
            PSF model to be applied. Presently must be `gaussian`, which models a single Gaussian.
            Will be extended in the future once TESS PRF models are made publicly available.
        xc: list, optional
            The x-coordinates of stars in the zeroth cadence. Must have length `nstars`. 
            While the positions of stars will be fit in all cadences, the relative positions of 
            stars will be fixed following the delta values from this list.
        yc: list, optional
            The y-coordinates of stars in the zeroth cadence. Must have length `nstars`. 
            While the positions of stars will be fit in all cadences, the relative positions of 
            stars will be fixed following the delta values from this list.
        """
        import tensorflow as tf
        from vaneska.models import Gaussian
        from tqdm import tqdm

        if len(xc) != nstars:
            raise ValueError('xc must have length nstars')
        if len(yc) != nstars:
            raise ValueError('yc must have length nstars')


        flux = tf.Variable(np.ones(nstars)*1000, dtype=tf.float64)
        bkg = tf.Variable(np.nanmedian(self.tpf[0]), dtype=tf.float64)
        xshift = tf.Variable(0.0, dtype=tf.float64)
        yshift = tf.Variable(0.0, dtype=tf.float64)

        if model == 'gaussian':

            gaussian = Gaussian(shape=self.tpf.shape[1:], col_ref=0, row_ref=0)

            a = tf.Variable(initial_value=1., dtype=tf.float64)
            b = tf.Variable(initial_value=0., dtype=tf.float64)
            c = tf.Variable(initial_value=1., dtype=tf.float64)

            if nstars == 1:
                mean = gaussian(flux, xc[0]+xshift, yc[0]+yshift, a, b, c)
            else:
                mean = [gaussian(flux[j], xc[j]+xshift, yc[j]+yshift, a, b, c) for j in range(nstars)]
        else:
            raise ValueError('This model is not incorporated yet!') # we probably want this to be a warning actually,
                                                                    # and a gentle return

        mean += bkg

        data = tf.placeholder(dtype=tf.float64, shape=self.tpf[0].shape)
        nll = tf.reduce_sum(tf.squared_difference(mean, data))

        var_list = [flux, xshift, yshift, a, b, c, bkg]
        grad = tf.gradients(nll, var_list)

        sess = tf.Session()
        sess.run(tf.global_variables_initializer())

        optimizer = tf.contrib.opt.ScipyOptimizerInterface(nll, var_list, method='TNC', tol=1e-4)

        fout = np.zeros((len(self.tpf), nstars))
        xout = np.zeros(len(self.tpf))
        yout = np.zeros(len(self.tpf))

        for i in tqdm(range(len(self.tpf))):
            optimizer.minimize(session=sess, feed_dict={data:self.tpf[i]}) # we could also pass a pointing model here
                                                                           # and just fit a single offset in all frames

            fout[i] = sess.run(flux)
            xout[i] = sess.run(xshift)
            yout[i] = sess.run(yshift)

        sess.close()

        self.psf_flux = fout[:,0]
        return


    def custom_aperture(self, shape=None, r=0.0, l=0.0, w=0.0, theta=0.0, pos=None, method='exact'):
        """
        Creates a custom circular or rectangular aperture of arbitrary size. 
        
        Parameters
        ----------       
        shape: str, optional
            The shape of the aperture to be used. Must be either `circle` or `rectangle.`
        r: float, optional
            If shape is `circle` the radius of the circular aperture to be used.
        l: float, optional
            If shape is `rectangle` the length of the rectangular aperture to be used.
        w: float, optional
            If shape is `rectangle` the width of the rectangular aperture to be used.
        theta: float, optional
            If shape is `rectangle` the rotation of the rectangle relative to detector coordinate. 
            Uses units of radians.   
        pos: tuple, optional
            The center of the aperture, in TPF coordinates. If not set, defaults to the center of the TPF.
        method: str, optional
            The method of producing a light curve to be used, either `exact`, `center`, or `subpixel`.
            Passed through to photutils and used as intended by that package.
        """
        if shape is None:
            print("Please select a shape: circle or rectangle")

        shape = shape.lower()
        if pos is None:
            pos = (self.tpf.shape[1]/2, self.tpf.shape[2]/2)
        else:
            pos = pos

        if shape == 'circle':
            if r == 0.0:
                print ("Please set a radius (in pixels) for your aperture")
            else:
                aperture = CircularAperture(pos, r=r)
                self.aperture = aperture.to_mask(method=method)[0].to_image(shape=((
                            np.shape(self.tpf[0]))))

        elif shape == 'rectangle':
            if l==0.0 or w==0.0:
                print("For a rectangular aperture, please set both length and width: custom_aperture(shape='rectangle', l=#, w=#)")
            else:
                aperture = RectangularAperture(pos, l=l, w=w, t=theta)
                self.aperture = aperture.to_mask(method=method)[0].to_image(shape=((
                            np.shape(self.tpf[0]))))
        else:
            print("Aperture shape not recognized. Please set shape == 'circle' or 'rectangle'")




    def jitter_corr(self, flux, cen=0.0):
        """
        Corrects for jitter in the light curve by quadratically regressing with centroid position. 
        Following Equation 1 of Knutson et al. 2008, ApJ, 673, 526. 
        
        Parameters
        ----------
        flux: numpy.ndarray 
            Time series of raw flux observations to be corrected.
        cen: float, optional 
            Center of the 2-d paraboloid for which the correction is performed. 
        """
        def parabola(params, x, y, f_obs, y_err):
            nonlocal cen
            c1, c2, c3, c4, c5 = params
            f_corr = f_obs * (c1 + c2*(x-cen) + c3*(x-cen)**2 + c4*(y-cen) + c5*(y-cen)**2)
            return np.sum( ((1-f_corr)/y_err)**2)


        # Masks out anything >= 2.5 sigma above the mean
        mask = np.ones(len(flux), dtype=bool)
        for i in range(5):
            lc_new = []
            std_mask = np.std(flux[mask])

            inds = np.where(flux <= np.mean(flux)-2.5*std_mask)
            y_err = np.ones(len(flux))**np.std(flux)
            for j in inds:
                y_err[j] = np.inf
                mask[j]  = False

            if i == 0:
                initGuess = [3, 3, 3, 3, 3]
            else:
                initGuess = test.x

            bnds = ((-15.,15.), (-15.,15.), (-15.,15.), (-15.,15.), (-15.,15.))
            centroid_xs, centroid_ys = self.centroid_xs-np.median(self.centroid_xs), self.centroid_ys-np.median(self.centroid_ys)
            test = minimize(parabola, initGuess, args=(centroid_xs, centroid_ys, flux, y_err), bounds=bnds)
            c1, c2, c3, c4, c5 = test.x
        lc_new = flux * (c1 + c2*(centroid_xs-cen) + c3*(centroid_xs-cen)**2 + c4*(centroid_ys-cen) + c5*(centroid_ys-cen)**2)
        return np.copy(lc_new)


    def set_header(self):
        """Defines the header for the TPF."""
        self.header = self.post_obj.header
        self.header.update({'CREATED':strftime('%Y-%m-%d')})

        # Removes postcard specific header information
        for keyword in ['POST_H', 'POST_W', 'CEN_X', 'CEN_Y', 'CEN_RA', 'CEN_DEC', 'POSTPIX1', 'POSTPIX2']:
            self.header.remove(keyword)

        # Adds TPF specific header information
        self.header.append(fits.Card(keyword='TIC_ID', value=self.source_info.tic,
                                     comment='TESS Input Catalog ID'))
        self.header.append(fits.Card(keyword='TMAG', value=self.source_info.tess_mag,
                                     comment='TESS mag'))
        self.header.append(fits.Card(keyword='GAIA_ID', value=self.source_info.gaia,
                                     comment='Associated Gaia ID'))
        self.header.append(fits.Card(keyword='SECTOR', value=self.source_info.sector,
                                     comment='Sector'))      
        self.header.append(fits.Card(keyword='CAMERA', value=self.source_info.camera,
                                     comment='Camera'))   
        self.header.append(fits.Card(keyword='CHIP', value=self.source_info.chip,
                                     comment='Chip'))                                  
        self.header.append(fits.Card(keyword='CHIPPOS1', value=self.source_info.position_on_chip[0],
                                     comment='central x pixel of TPF in FFI chip'))
        self.header.append(fits.Card(keyword='CHIPPOS2', value=self.source_info.position_on_chip[1],
                                     comment='central y pixel of TPF in FFI'))
        self.header.append(fits.Card(keyword='POSTCARD', value=self.source_info.postcard,
                                     comment='Postcard'))        
        self.header.append(fits.Card(keyword='POSTPOS1', value= self.source_info.position_on_postcard[0],
                                     comment='predicted x pixel of source on postcard'))
        self.header.append(fits.Card(keyword='POSTPOS2', value= self.source_info.position_on_postcard[1],
                                     comment='predicted y pixel of source on postcard'))
        self.header.append(fits.Card(keyword='CEN_RA', value = self.source_info.coords[0],
                                     comment='RA of TPF source'))
        self.header.append(fits.Card(keyword='CEN_DEC', value=self.source_info.coords[1],
                                     comment='DEC of TPF source'))
        self.header.append(fits.Card(keyword='TPF_H', value=np.shape(self.tpf[0])[0],
                                     comment='Height of the TPF in pixels'))
        self.header.append(fits.Card(keyword='TPF_W', value=np.shape(self.tpf[0])[1],
                                           comment='Width of the TPF in pixels'))



    def save(self, output_fn=None, directory=None):
        """Saves a created TPF object to a FITS file.

        Parameters
        ----------
        output_fn : str, optional
            Filename to save output as. Overrides default naming.
        directory : str, optional
            Directory to save file into.
        """

        # if the user did not specify a directory, set it to default
        if directory is None:
            directory = self.fetch_dir()

        # Creates column names for FITS tables
        r = np.arange(1.5,4,0.5)
        colnames=[]
        for i in r:
            colnames.append('_'.join(str(e) for e in ['circle', 'binary', i]))
            colnames.append('_'.join(str(e) for e in ['rectangle', 'binary', i]))
            colnames.append('_'.join(str(e) for e in ['circle', 'weighted', i]))
            colnames.append('_'.join(str(e) for e in ['rectangle', 'weighted', i]))
        raw       = [e+'_raw' for e in colnames]
        errors    = [e+'_err' for e in colnames]
        corrected = [e+'_corr' for e in colnames]

        # Creates table for first extension (tpf, tpf_err, best lc, time, centroids)
        ext1 = Table()
        ext1['TIME']       = self.time
        ext1['TPF']        = self.tpf
#        ext1['FLUX']       = self.tpf
        ext1['TPF_ERR']    = self.tpf_err
        ext1['RAW_FLUX']   = self.raw_flux
        ext1['CORR_FLUX']  = self.corr_flux
        ext1['FLUX_ERR']   = self.flux_err
        ext1['QUALITY']    = self.quality
        ext1['X_CENTROID'] = self.centroid_xs
        ext1['Y_CENTROID'] = self.centroid_ys
        ext1['X_COM']      = self.x_com
        ext1['Y_COM']      = self.y_com

        # Creates table for second extension (all apertures)
        ext2 = Table()
        for i in range(len(self.all_apertures)):
            ext2[colnames[i]] = self.all_apertures[i]

        # Appends custom aperture to the end, if there is one
        if self.custom_aperture is not None:
            ext2['custom'] = self.custom_aperture

        # Creates table for third extention (all raw & corrected fluxes and errors)
        ext3 = Table()
        for i in range(len(raw)):
            ext3[raw[i]]       = self.all_raw_lc[i]
            ext3[corrected[i]] = self.all_corr_lc[i]
            ext3[errors[i]]    = self.all_lc_err[i]

        # Appends aperture to header
        self.header.append(fits.Card(keyword='APERTURE', value=colnames[self.best_ind],
                                     comment='Best aperture used for lightcurves in extension 1'))

        # Writes TPF to FITS file
        primary_hdu = fits.PrimaryHDU(header=self.header)
        data_list = [primary_hdu, fits.BinTableHDU(ext1), fits.BinTableHDU(ext2), fits.BinTableHDU(ext3)]
        hdu = fits.HDUList(data_list)

        if output_fn==None:
            hdu.writeto(os.path.join(directory,
                        'hlsp_eleanor_tess_ffi_lc_TIC{}_s{}_v0.1.fits'.format(
                        self.source_info.tic, self.source_info.sector),
                        overwrite=True))
        else:
            hdu.writeto(output_fn)



    def load(self, directory=None):
        """
        Loads in and sets all the attributes for a pre-created TPF file.

        Parameters
        ----------
        directory : str, optional
            Directory to load file from.
        """

        if directory is None:
            directory = self.fetch_dir()

        hdu = fits.open(os.path.join(directory, self.source_info.fn))
        hdr = hdu[0].header
        self.header = hdr
        # Loads in everything from the first extension
        cols  = hdu[1].columns.names
        table = hdu[1].data
        self.time        = table[cols[0]]
        self.tpf         = table[cols[1]]
        self.tpf_err     = table[cols[2]]
        self.raw_flux    = table[cols[3]]
        self.corr_flux   = table[cols[4]]
        self.flux_err    = table[cols[5]]
        self.quality     = table[cols[6]]
        self.centroid_xs = table[cols[7]]
        self.centroid_ys = table[cols[8]]

        # Loads in apertures from second extension
        self.all_apertures = []
        cols  = hdu[2].columns.names
        table = hdu[2].data
        for i in cols:
            if i == 'custom':
                self.custom_aperture = table[i]
            elif i == hdr['aperture']:
                self.aperture = table[i]
            else:
                self.all_apertures.append(table[i])

        # Loads in remaining light curves from third extension
        cols  = hdu[3].columns.names
        table = hdu[3].data
        self.all_raw_lc  = []
        self.all_corr_lc = []
        self.all_lc_err  = []
        for i in cols:
            if i[-4::] == 'corr':
                self.all_corr_lc.append(table[i])
            elif i[-3::] == 'err':
                self.all_lc_err.append(table[i])
            else:
                self.all_raw_lc.append(table[i])
        return

    def fetch_dir(self):
        """
        Returns the default path to the directory where files will be saved
        or loaded.

        By default, this method will return "~/.eleanor" and create
        this directory if it does not exist.  If the directory cannot be
        access or created, then it returns the local directory (".").

        Returns
        -------
        download_dir : str
            Path to location of `ffi_dir` where FFIs will be downloaded
        """
        download_dir = os.path.join(os.path.expanduser('~'), '.eleanor')
        if os.path.isdir(download_dir):
            return download_dir
        else:
            # if it doesn't exist, make a new cache directory
            try:
                os.mkdir(download_dir)
            # downloads locally if OS error occurs
            except OSError:
                warnings.warn('Warning: unable to create {}. '
                              'Downloading FFIs to the current '
                              'working directory instead.'.format(download_dir))
                download_dir = '.'

        return download_dir



import numpy as np
from astropy.wcs import WCS
from astropy.table import Table
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
import sys

import urllib
from .mast import *
from .utils import *

__all__ = ['Source', 'multi_sectors']


def multi_sectors(sectors, tic=None, gaia=None, coords=None):
    """Obtain a list of Source objects for a single target, for each of multiple sectors for which the target was observed.
    
    Parameters
    ----------
    tic : int, optional
        The TIC ID of the source.
    gaia : int, optional
        The Gaia DR2 source_id.
    coords : tuple, optional
        The (RA, Dec) coords of the object in degrees.
    sectors : list or str
        The list of sectors for which data should be returned, or `all` to return all sectors 
        for which there are data.
    """
    objs = []
    if sectors == 'all':
<<<<<<< HEAD
        sectors = list(np.arange(1,14,1, dtype=int))
=======
        sectors = list(np.arange(1,14,1))
>>>>>>> 081771d1d2cccd4265c477102654d6a3c61fda72
    if type(sectors) == list:
        for s in sectors:
            star = Source(tic=tic, gaia=gaia, coords=coords, sector=int(s))
            if star.sector is not None:
                objs.append(star)
        return objs
    else:
        raise TypeError("Sectors needs to be either 'all' or a type(list) to work.")
    

def load_postcard_guide():
    """Load and return the postcard coordinates guide."""
    guide_link = urllib.request.urlopen('http://astro.uchicago.edu/~bmontet/TESS_postcards/postcard.guide')
    guide = guide_link.read().decode('utf-8')
    guide = Table.read(guide, format='ascii.basic') # guide to postcard locations
    return guide

class Source(object):
    """A single source observed by TESS.

    Parameters
    ----------
    tic : int, optional
        The TIC ID of the source.
    gaia : int, optional
        The Gaia DR2 source_id.
    coords : tuple or astropy.coordinates.SkyCoord, optional
        The (RA, Dec) coords of the object in degrees or an astropy SkyCoord object.
    fn : str, optional
        Filename of a TPF corresponding to the desired source.
    sector : int or str 
        The sector for which data should be returned, or `recent` to 
        obtain data for the most recent sector which contains this target. 

    Attributes
    ----------
    tess_mag : float
        The TESS magnitude from the TIC.
    sector : int
        Sector in which source was observed by TESS.
    camera : int
        TESS camera on which source falls.
    chip : int
        TESS chip on which source falls.
    position_on_chip : (int, int)
        Predicted (x,y) coords of object center on chip.
    postcard : str
        Name of the best postcard (postcard where source should be located closest to the center).
    position_on_postcard : (int, int)
        Predicted (x, y) coords of object center on the above-named postcard.
        Does NOT take into account additional pointing corrections; these will be applied when making the TPF.
    all_postcards : list of strs
        Names of all postcards where the source appears.
    """
    def __init__(self, tic=None, gaia=None, coords=None, fn=None, sector=None):
        self.tic     = tic
        self.gaia    = gaia
        self.coords  = coords
        self.fn      = fn
        self.premade = False
        self.usr_sec = sector

        if self.fn is not None:
            try:
                hdu = fits.open(self.fn)
            except:
                assert False, "{0} is not a valid filename.".format(fn)
            hdr = hdu[0].header
            self.tic      = hdr['TIC_ID']
            self.tess_mag = hdr['TMAG']
            self.gaia     = hdr['GAIA_ID']
            self.coords   = (hdr['CEN_RA'], hdr['CEN_DEC'])
            self.premade  = True
            self.sector   = hdr['SECTOR']
            self.camera   = hdr['CAMERA']
            self.chip     = hdr['CHIP']
            self.position_on_chip = (hdr['CHIPPOS1'], hdr['CHIPPOS2'])
            self.position_on_postcard = (hdr['POSTPOS1'], hdr['POSTPOS2'])
                    
        else:
            if self.coords is not None:
                assert type(self.coords) is astropy.coordinates.sky_coordinate.SkyCoord, \
                    "Source: coords must be an astropy.coordinates.SkyCoord object."
                if type(self.coords) is astropy.coordinates.sky_coordinate.SkyCoord:
                    self.coords = (c.ra.degree, c.dec.degree)
                elif (len(coords) == 2) & all(isinstance(c, float) for c in self.coords):
                    self.coords  = coords
                else:
                    assert False, ("Source: invalid coords. Valid input types are: "
                                   "(RA, Dec) tuple or astropy.coordinates.SkyCoord object.")

                self.tic, self.tess_mag, sep = tic_from_coords(self.coords)
                self.gaia = gaia_from_coords(self.coords)

            elif self.gaia is not None:
                self.coords = coords_from_gaia(self.gaia)
                self.tic, self.tess_mag = tic_from_coords(self.coords)

            elif self.tic is not None:
                self.coords, self.tess_mag = coords_from_tic(self.tic)
                self.gaia = gaia_from_coords(self.coords)
                
            else:
                assert False, ("Source: one of the following keywords must be given: "
                               "tic, gaia, coords, fn.")

            self.tess_mag = self.tess_mag[0]
            self.locate_on_tess() # sets sector, camera, chip, postcard,
                                  # position_on_chip, position_on_postcard



    def locate_on_chip(self, guide):
        """Finds the TESS sector, camera, chip, and position on chip for the source.
        Sets attributes sector, camera, chip, position_on_chip.
        """
        def cam_chip_loop(sec):
            for cam in np.unique(guide['CAMERA']):
                for chip in np.unique(guide['CCD']):
                    mask = ((guide['SECTOR'] == sec) & (guide['CAMERA'] == cam)
                            & (guide['CCD'] == chip))
                    if np.sum(mask) == 0:
                        continue # this camera-chip combo is not in the postcard database yet
                    else:
                        random_postcard = guide[mask][0]
                    d = {}
                    for j,col in enumerate(random_postcard.colnames):
                        d[col] = random_postcard[j]
                    hdr = fits.Header(cards=d) # make WCS info from one postcard header
                    xy = WCS(hdr).all_world2pix(self.coords[0], self.coords[1], 1, quiet=True) # position in pixels in FFI dims
                    x_zero, y_zero = hdr['POSTPIX1'], hdr['POSTPIX2']
                    xy = np.array([xy[0]+x_zero, xy[1]+y_zero])
                    if (0 <= xy[0] < 2048) & (44 <= xy[1] < 2092):
                        self.sector = sec
                        self.camera = cam
                        self.chip = chip
                        self.position_on_chip = np.ravel(xy)

        guide = load_postcard_guide()
        self.sector=None

        if self.usr_sec is None:
            for sec in np.unique(guide['SECTOR']):
                cam_chip_loop(sec)
            
            if self.sector is None:
                raise SearchError("TESS has not (yet) observed your target.")

        elif self.usr_sec is not None:
            if type(self.usr_sec) == int:
                cam_chip_loop(self.usr_sec)

            # Searches for the most recent sector the object was observed in
            elif self.usr_sec.lower() == 'recent':
                for s in np.arange(15,0,-1):
                    cam_chip_loop(s)
                    if self.sector is not None:
                        break

        return


    def locate_on_tess(self):
        """Finds the best TESS postcard(s) and the position of the source on postcard.
        Sets attributes postcard, position_on_postcard, all_postcards.
        """
        guide = load_postcard_guide()
        self.locate_on_chip(guide)

        if self.sector is None:
            return

        # Searches through postcards for the given sector, camera, chip
        in_file, dists=[], []

        # Searches through rows of the table
        for i in range(len(guide)):
            postcard_inds = np.arange(len(guide))[(guide['SECTOR'] == self.sector) & (guide['CAMERA'] == self.camera)
                                                  & (guide['CCD'] == self.chip)]
        xy = self.position_on_chip

        # Finds postcards containing the source
        for i in postcard_inds: # loop rows

            x_cen, y_cen= guide['CEN_X'][i], guide['CEN_Y'][i]
            l, w = guide['POST_H'][i]/2., guide['POST_W'][i]/2.

            # Checks to see if xy coordinates of source falls within postcard
            if (xy[0] >= x_cen-l) & (xy[0] <= x_cen+l) & (xy[1] >= y_cen-w) & (xy[1] <= y_cen+w):
                in_file.append(i)
                dists.append(np.min([xy[0]-(x_cen-l), (x_cen+l)-xy[0], xy[1]-(y_cen-w), (y_cen+w)-xy[1]]))


        # If more than one postcard is found for a single source, choose the postcard where the
        # source is closer to the center
        if in_file == []:
            raise SearchError("Sorry! We don't have a postcard for you at the moment.")
        elif len(in_file) > 1:
            best_ind = np.argmax(dists)
        else:
            best_ind = 0

        self.all_postcards = guide['POSTNAME'][in_file]
        self.postcard = guide['POSTNAME'][in_file[best_ind]]

        self.sector = guide['SECTOR'][in_file[best_ind]] # WILL BREAK FOR MULTI-SECTOR TARGETS
        self.camera = guide['CAMERA'][in_file[best_ind]]
        self.chip = guide['CCD'][in_file[best_ind]]

        i = in_file[best_ind]
        postcard_pos_on_ffi = (guide['CEN_X'][i] - guide['POST_H'][i]/2.,
                                guide['CEN_Y'][i] - guide['POST_W'][i]/2.)
        self.position_on_postcard = xy - postcard_pos_on_ffi # as accurate as FFI WCS

        i = in_file[best_ind]
        postcard_pos_on_ffi = (guide['CEN_X'][i] - guide['POST_H'][i]/2.,
                              guide['CEN_Y'][i] - guide['POST_W'][i]/2.) # (x,y) on FFI where postcard (x,y) = (0,0)
        self.position_on_postcard = xy - postcard_pos_on_ffi # as accurate as FFI WCS
